from .commands import command_run, command_makemigrations
from .db import db_queries
from .live import liveserver, liveclient
from .mail import mail_outbox
from .messages import messages
from .migrations import check_migrations
from .request import request_factory, request_get, request_post, request_client
from .settings import settings
from .templates import template_render_tag, template_context, template_strip_tags
from .users import user_create, user_model, user
from .utils import conf_app_name

fixtures_registered: bool = True



from typing import List

import pytest
from decimal import Decimal
from django.db import connections, DEFAULT_DB_ALIAS, reset_queries

if False:  # pragma: nocover
    from collections import deque  # noqa


@pytest.fixture
def db_queries(settings) -> 'Queries':
    """Allows access to executed DB queries.

    Example::

        def test_db(db_queries):

            # Previous queries cleared at the beginning.
            assert len(db_queries) == 0

            ...  # Do some DB-related stuff.

            # Assert total queries on all DBs.
            assert len(db_queries) == 10

            # Default DBs SQLs with auxiliary commands filtered out by default.
            sqls = db_queries.sql()

            # Assert total execution time is less than a second.
            assert db_queries.time() < 1

            # Drop SQL gathered so far on default DB.
            db_queries.clear()


    .. warning:: Requires Django 1.9+ to work.

    """
    queries = Queries()

    debug_values_prev = {}

    for connection in connections.all():
        debug_values_prev[connection.alias] = connection.force_debug_cursor
        connection.force_debug_cursor = True

    try:
        queries.clear_all()

        yield queries

    finally:

        for connection in connections.all():
            prev_debug_value = debug_values_prev.get(connection.alias, None)

            if prev_debug_value is not None:
                connection.force_debug_cursor = prev_debug_value

        queries.clear_all()


class Queries:
    """Allows access to executed DB queries."""

    sql_drop = {
        'BEGIN',
        'COMMIT',
        'END',
    }

    def __len__(self) -> int:
        return len(self.get_log())

    def get_log(self, db_alias: str = None) -> 'deque':
        """
        :param db_alias:

        """
        return connections[db_alias or DEFAULT_DB_ALIAS].queries_log

    def clear_all(self):
        """Clears all queries logged for all DBs."""
        reset_queries()

    def clear(self, db_alias: str = None):
        """Clear queries for the given or default DB.

        :param db_alias: Database alias. Default is used if not given.

        """
        self.get_log(db_alias=db_alias).clear()

    def sql(self, db_alias: str = None, *, drop_auxiliary: bool = True) -> List[str]:
        """Returns a list of queries executed using the given or default DB.

        :param db_alias: Database alias. Default is used if not given.

        :param drop_auxiliary: Filter out auxiliary SQL like:
            * BEGIN
            * COMMIT
            * END

        """
        sqls = []

        auxiliary = self.sql_drop

        for log_entry in self.get_log(db_alias=db_alias):
            sql = ' '.join(sql_line.strip() for sql_line in log_entry['sql'].splitlines())
            if not drop_auxiliary or sql not in auxiliary:
                sqls.append(sql)

        return sqls

    def time(self, db_alias: str = None) -> Decimal:
        """Returns total time executing queries (in seconds) using the given or default DB.

        :param db_alias: Database alias. Default is used if not given.

        """
        times = [Decimal(log_entry['time']) for log_entry in self.get_log(db_alias=db_alias)]
        return sum(times)



try:
    from django.urls import re_path
    
except ImportError:
    from django.conf.urls import url as re_path

    
from django.http import HttpResponse

from pytest_djangoapp.compat import get_urlpatterns


def raise_exception(request):
    raise Exception('This one should be handled by 500 technical view')


def index(request, some_id):

    if request.method == 'POST':
        return HttpResponse('json')

    if request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest':  # .headers in Django 2.2+
        return HttpResponse('ajaxed')

    from django.templatetags.static import static
    return HttpResponse(f"{some_id} | fine {static('blank.png')}")


urlpatterns = get_urlpatterns([
    re_path(r'^index/(?P<some_id>\d+)/$', index, name='index'),
    re_path(r'^raiser/$', raise_exception, name='raiser'),
])



from django.core.management.base import BaseCommand

from .compat import options_getter, CommandOption

get_options = options_getter((
    CommandOption('--two', action='store'),
))


class Command(BaseCommand):

    help = 'Sends scheduled messages (both in pending and error statuses).'

    option_list = get_options()

    def add_arguments(self, parser):
        parser.add_argument('one')
        get_options(parser.add_argument)

    def handle(self, one, **options):
        self.stdout.write('bingo')
        self.stderr.write('bongo')

        two = options.get('two', None)

        return f'{one}|{two}'



def test_request_factory(request_factory):
    assert request_factory()


def test_request_get(request_get, user_create):
    assert request_get(user=user_create())
    assert request_get(ajax=True)


def test_request_post(request_post, user_create):
    assert request_post(data={'a': 'b'}, user=user_create())
    assert request_post(ajax=True, data={'a': 'b'})


def test_request_client(request_client, user_create):
    client = request_client()
    assert client
    assert client.user is None
    assert not client.user_logged_in

    response = client.get(('index', {'some_id': 10}))

    assert response.status_code == 200
    assert response.content == b'10 | fine /static/blank.png'

    # Now AJAX client.
    client = request_client(ajax=True)

    response = client.get(('index', {'some_id': 22}))

    assert response.status_code == 200
    assert b'ajaxed' in response.content

    new_user = user_create()
    client = request_client(user=new_user)

    assert client.user is new_user
    assert client.user_logged_in

    # now technical 500 view
    client = request_client(raise_exceptions=False)
    response = client.get('/raiser/')
    assert b'<h1>Server Error (500)</h1>' in response.content

    # now json
    client = request_client(json=True)
    response = client.post(('index', {'some_id': 10}), {'a': 'b'})
    content = response.content
    assert b'json' in content



class CommandOption:
    """Command line option wrapper."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


def options_getter(command_options):
    """Compatibility function to get rid of optparse in management commands after Django 1.10.

    :param tuple command_options: tuple with `CommandOption` objects.

    """
    def get_options(option_func=None):
        from optparse import make_option

        func = option_func or make_option

        options = tuple([func(*option.args, **option.kwargs) for option in command_options])

        if option_func is None:
            result = []

        else:
            result = options

        return result

    return get_options




def test_user_create(user_create):

    assert not user_create().is_superuser
    assert user_create(superuser=True).is_superuser
    assert user_create(anonymous=True).is_anonymous


def test_user_model(user_model):

    assert user_model


def test_user(user):

    assert user
    assert user.password_plain == 'password'



import pytest

from pytest_djangoapp.tests.testapp.models import Article
from django import VERSION


def test_models_basic():
    article = Article(title='my title')
    article.save()


@pytest.mark.skipif(VERSION < (1, 9), reason='Django 1.9+ required')
def test_db_queries(db_queries):

    assert len(db_queries) == 0

    article = Article(title='my title')
    article.save()

    print(db_queries.get_log())

    assert len(db_queries) == 1 if VERSION >= (2, 2) else 2

    assert len(db_queries.sql()) == 1  # auxiliary sql filtered
    assert db_queries.time() < 1

    db_queries.clear()



from typing import List

import pytest

from django.core import mail


mail.outbox = []


@pytest.fixture()
def mail_outbox() -> List[mail.EmailMessage]:
    """Returns mail outbox: list of sent message objects.

    Example::

        def test_this(mail_outbox):
            first_subject = mail_outbox[0].subject

    """
    yield mail.outbox

    mail.outbox = []



from typing import Iterable, Optional

import pytest

from django.core.management import call_command


@pytest.fixture
def command_run():
    """Allows management command run.

    Example::

        def test_this(command_run, capsys):
            result = command_run('my_command', args=['one'], options={'two': 'three'})
            out, err = capsys.readouterr()


    .. warning:: Django < 1.10 will always return `None`, no matter what command returns.

    :param command_name: Command name to run.
    :param args: Required arguments to pass to a command.
    :param options: Optional arguments to pass to a command.

    :returns: Command output.

    """
    def command_run_(command_name: str, args: Iterable[str] = None, options: dict = None) -> Optional[str]:
        args = args or []
        options = options or {}
        return call_command(command_name, *args, **options)

    return command_run_


@pytest.fixture
def command_makemigrations(conf_app_name):
    """Allows to run makemigrations command.

    .. note:: This command can be useful to generate
        migrations for your application (without a project creation).

    Example::

        def test_makemigrations(command_makemigrations):
            command_makemigrations()

    :param app: Application name to run 'makemigrations' for.
        * By default, a name from 'conf_app_name' fixture is used.
        * If empty string, command is run for any application.

    :param args: Additional arguments to pass to the command.

    """
    def command_migration_(*, app: str = None, args: Iterable[str] = None) -> Optional[str]:
        args = args or []

        if app is None:
            app = conf_app_name

        if app:
            args.append(app)

        return call_command('makemigrations', *args)

    return command_migration_



from pytest_djangoapp import configure_djangoapp_plugin


def hook(settings):
    settings['HOOKED'] = True
    return settings


pytest_plugins = configure_djangoapp_plugin(

    admin_contrib=True,

    extend_INSTALLED_APPS=[
        'django.contrib.sites',
    ],

    extend_MIDDLEWARE=[
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',  # Bogus doubled check.
    ],

    extend_DATABASES={
        'dummy': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:',
        },
    },

    settings_hook=hook,
    migrate=True,

)



import pytest

from pytest_djangoapp.configuration import Configuration


def test_configration(pytestconfig):

    settings = Configuration.get()
    settings[Configuration.KEY_APP] = 'some'

    assert Configuration.get_combined(pytestconfig)

    settings[Configuration.KEY_APP] = ''
    assert Configuration.get_combined(pytestconfig)

    def swap_dir(level):

        old_dir = pytestconfig.invocation_dir

        try:
            pytestconfig.invocation_dir = old_dir.parts()[level]
            assert Configuration.get_combined(pytestconfig)

        finally:
            pytestconfig.invocation_dir = old_dir

    with pytest.raises(Exception):
        # Unable to deduce app name.
        swap_dir(-4)


def test_settings_hook():
    from django.conf import settings
    assert settings.HOOKED



from typing import Type, TypeVar, Dict

import pytest
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

TypeLiveClient = TypeVar('TypeLiveClient', bound='LiveClient')


@pytest.fixture()
def liveserver() -> Type['LiveServer']:
    """Runs a live server. Available as a context manager.

    .. warning:: For Django >= 4.0

    Example::

        def test_live(liveserver):

            with liveserver() as server:
                print(f'Live server is available at: {server.url}')

    """
    return LiveServer


@pytest.fixture()
def liveclient():
    """Runs a live client. Available as a context manager.


    Example::

        def test_live(liveserver, liveclient):

            with liveserver() as server:

                # Let's run Firefox using Selenium.
                with liveclient('selenium', browser='firefox') as client:
                    selenium = client.handle  # Selenium driver is available in .handle

                    # Let's open server's root URL and check heading 1 on that page
                    selenium.get(server.url)
                    assert selenium.find_element('tag name', 'h1').text == 'Not Found'

    """

    def get_client(typename: str, *, browser: str) -> TypeLiveClient:
        return LiveClient.spawn(alias=typename, browser=browser)

    return get_client


class LiveServer:

    _cls = StaticLiveServerTestCase

    def __init__(self, *, host: str = None, port: int = None):

        cls = self._cls

        if host:
            cls.host = host

        if port is not None:
            cls.port = port

    @property
    def url(self) -> str:
        """URL to access this live server."""
        return self._cls.live_server_url

    def __enter__(self):
        self._cls._start_server_thread()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cls._terminate_thread()


class LiveClient:
    """Base class for live clients."""

    alias: str = ''

    _registry: Dict[str, TypeLiveClient] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()

        alias = cls.alias

        if alias:
            cls._registry[alias] = cls

    def __init__(self, *, browser: str):
        self._browser = browser
        self.handle = None

    def __enter__(self) -> TypeLiveClient:
        self.handle = self._handle_init()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._handle_destroy()

    @classmethod
    def spawn(cls, alias: str, **kwargs) -> TypeLiveClient:
        client_cls = LiveClient._registry[alias]
        return client_cls(**kwargs)

    def _handle_init(self):  # pragma: nocover
        raise NotImplementedError

    def _handle_destroy(self):
        pass


class SeleniumLiveClient(LiveClient):
    """This live client wraps Selenium.

    https://selenium-python.readthedocs.io/

    """
    alias: str = 'selenium'

    def _handle_init(self):
        from django.test.selenium import SeleniumTestCaseBase

        cls = SeleniumTestCaseBase
        browser = self._browser
        driver_cls = cls.import_webdriver(browser)

        return driver_cls(options=cls.import_options(browser)())

    def _handle_destroy(self):

        handle = self.handle

        if handle:
            handle.quit()

        super()._handle_destroy()



from time import time

import pytest

from ..configuration import Configuration


def get_stamp() -> str:
    """Returns current timestamp as a string."""
    return f'{time()}'


@pytest.fixture(scope='session')
def conf_app_name() -> str:
    """Returns your application name (deduced or provided)."""
    return Configuration.get_dict()[Configuration.KEY_APP]



from django import VERSION


def test_command_run(command_run, capsys):

    result = command_run('my_command', args=['one'], options={'two': 'three'})
    out, err = capsys.readouterr()

    assert 'bingo' in out
    assert 'bongo' in err

    if VERSION >= (1, 10):
        assert result == 'one|three'


def test_command_migration(command_makemigrations, capsys):

    result = command_makemigrations()
    out, err = capsys.readouterr()

    assert result is None

    assert 'No changes' in out
    assert err == ''



from django.conf import settings as django_settings
from django.core.management import call_command
from django.test.runner import DiscoverRunner

from .fixtures import *  # noqa
from .toolbox import Configuration

runner = DiscoverRunner(interactive=False, verbosity=0)
setup_databases = runner.setup_databases
teardown_databases = runner.teardown_databases


def pytest_runtest_setup(item):
    # todo optional db initialization
    item.old_config = setup_databases()


def pytest_runtest_teardown(item, nextitem):
    unset = {}

    old_config = getattr(item, 'old_config', unset)

    if old_config is unset:
        # _teardown will suddenly work even if no _setup has occured
        # e.g. in case of mark.skipif
        return

    call_command('flush', interactive=False, reset_sequences=False)
    teardown_databases(old_config)


def pytest_configure(config):
    import django

    settings_dict = Configuration.get_combined(config)

    django_settings.configure(**settings_dict)
    django.setup()
    runner.verbosity = config.option.verbose



from typing import Union

import pytest

from .utils import get_stamp

if False:  # pragma: nocover
    from django.contrib.auth.models import AbstractUser, AnonymousUser  # noqa


TypeUser = Union['AbstractUser', 'AnonymousUser']


@pytest.fixture()
def user_model() -> TypeUser:
    """Returns user model class.

    Example::

        def test_this(user_model):
            model_cls = user_model

    """
    from django.contrib.auth import get_user_model

    return get_user_model()


@pytest.fixture()
def user_create(user_model):
    """Allows Django user object generation.

    Example::

        def test_this(user_create):
            user = user_create()

    .. note:: User password is accessible via `password_plain` attribute.

    :param superuser: Whether to create a superuser.
    :param anonymous: Whether to create an anonymous user.
    :param attributes: Additional user object attributes to initialize.

    """
    from django.contrib.auth.models import AnonymousUser

    def user_create_(*, superuser: bool = False, anonymous: bool = False, attributes: dict = None) -> TypeUser:

        if anonymous:
            return AnonymousUser()

        attributes = attributes or {}
        manager = user_model._default_manager

        username = (('admin' if superuser else 'user') + '_' + get_stamp())

        kwargs = {
            'username': username,
            'email': f'{username}@example.com',
            'password': 'password',
        }

        kwargs.update(attributes)

        if superuser:
            user = manager.create_superuser(**kwargs)
        else:
            user = manager.create_user(**kwargs)

        # Keep plain password for test purposes.
        user.password_plain = kwargs['password']

        return user

    return user_create_


@pytest.fixture()
def user(user_create) -> TypeUser:
    """Exposes Django user object.

    Shortcut for `user_create` fixture.

    Example::

        def test_this(user):
            username = user.username

    .. note:: User password is accessible via `password_plain` attribute.

    """
    return user_create()



from threading import local
from typing import Callable

_THREAD_LOCAL = local()
setattr(_THREAD_LOCAL, 'configuration', {})


class FakeMigrationModules:
    """Allows skipping migration applying process."""

    def __init__(self, module_name: str):
        self.module_name = module_name

    def __getitem__(self, item: str) -> str:
        return self.module_name

    def __contains__(self, item: str):
        return True


class Configuration:

    _prefix = 'DJANGOAPP_OPTIONS'
    _KEY_ADMIN = 'admin'
    KEY_APP = 'app_name'
    _KEY_EXTEND = 'extend'
    _KEY_HOOK = 'hook'
    _KEY_MIGRATE = 'migrate'

    DIR_TESTAPP = 'testapp'
    """Name of test application directory.
    
    Test application directory should be placed inside `tests` directory 
    and needs to be a Python package (contain __init__.py).
     
    Test application is useful to place there modules like `urls.py`,
    `models.py` (e.g. with custom models), etc.
    
    """

    @classmethod
    def get(cls) -> dict:
        return _THREAD_LOCAL.configuration

    @classmethod
    def set(
        cls,
        settings_dict: dict = None,
        *,
        app_name: str = None,
        admin_contrib: bool = False,
        settings_hook: Callable = None,
        migrate: bool = True,
        **kwargs
    ):
        """
        :param settings_dict:

        :param app_name:

        :param admin_contrib: Setup Django to test Admin contrib related parts.

        :param settings_hook: Allows setting a function to get resulting settings.

            Function must accept settings dictionary, and return resulting settings dictionary.

            .. code-block:: python

                def hook_func(settings):
                    return settings

        :param migrate: Allows applying or skipping migration applying process.
            Skipping could be useful for testing applications with many migrations.

        :param kwargs: Additional arguments.

            Use `extend_` prefix to extend default configuration.
            E.g.: extend_INSTALLED_APPS=['a']

        """
        settings_dict = settings_dict or {}

        extend = {}

        for key, val in kwargs.items():
            _, _, extend_key = key.partition('extend_')

            if extend_key and extend_key == extend_key.upper():
                extend[extend_key] = val

        base_settings = {
            cls._prefix: {
                cls.KEY_APP: app_name,
                cls._KEY_EXTEND: extend,
                cls._KEY_ADMIN: admin_contrib,
                cls._KEY_HOOK: settings_hook,
                cls._KEY_MIGRATE: migrate,
            }
        }

        base_settings.update(settings_dict)

        _THREAD_LOCAL.configuration = base_settings

    @classmethod
    def get_defaults(cls) -> dict:
        from django.conf import global_settings

        if hasattr(global_settings, 'MIDDLEWARE_CLASSES'):
            middleware = global_settings.MIDDLEWARE_CLASSES

        else:
            middleware = global_settings.MIDDLEWARE

        installed_apps = list(global_settings.INSTALLED_APPS[:])
        installed_apps.extend([
            'django.contrib.auth',
            'django.contrib.contenttypes',
        ])
        installed_apps = list(set(installed_apps))

        settings_dict = dict(

            SECRET_KEY='djangoapp',

            ALLOWED_HOSTS=(
                global_settings.ALLOWED_HOSTS +
                # Satisfy Django test client needed in Django < 2.0
                ['testserver']
            ),

            INSTALLED_APPS=installed_apps,
            STATIC_URL='/static/',

            DATABASES={
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': ':memory:',
                },
            },

            MIDDLEWARE=middleware,

            EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend',

            TEMPLATES=[
                {
                    'BACKEND': 'django.template.backends.django.DjangoTemplates',
                    'APP_DIRS': True,
                    'OPTIONS': {'context_processors': ['django.contrib.messages.context_processors.messages']}
                },
            ],

        )

        return settings_dict.copy()

    @classmethod
    def get_combined(cls, pytest_config) -> dict:
        from django import VERSION

        settings = cls.get()

        defaults = cls.get_defaults()
        defaults.update(settings)

        djapp_options = defaults[cls._prefix]

        app_name = djapp_options[cls.KEY_APP]
        extensions = djapp_options[cls._KEY_EXTEND]
        admin = djapp_options[cls._KEY_ADMIN]
        hook = djapp_options.pop(cls._KEY_HOOK, None) or (lambda settings_dict: settings_dict)

        # djangoapp is considered testing a whole project (a set of apps)
        # if hook function is a `partial` for function with a certain name.
        project_mode = getattr(getattr(hook, 'func', None), '__name__', '') == 'update_settings_from_module'

        if not djapp_options[cls._KEY_MIGRATE]:
            module_name = None

            if VERSION <= (1, 10):
                module_name = 'dummy_migrations'
                defaults['MIGRATIONS_MODULE_NAME'] = module_name

            defaults['MIGRATION_MODULES'] = FakeMigrationModules(module_name)

        if admin:
            middleware = extensions.setdefault('MIDDLEWARE', [])
            middleware.extend([
                'django.contrib.sessions.middleware.SessionMiddleware',
                'django.contrib.auth.middleware.AuthenticationMiddleware',
            ])
            apps = extensions.setdefault('INSTALLED_APPS', [])
            apps.extend([
                'django.contrib.admin',
                'django.contrib.sessions',
            ])

        for key, value in extensions.items():
            default_value = defaults.get(key, [])

            if isinstance(default_value, (list, tuple)):
                extended = list(default_value)

                for item in value:
                    if item not in extended:
                        extended.append(item)

                defaults[key] = extended

                if key == 'MIDDLEWARE':
                    # Special case for renamed.
                    defaults['MIDDLEWARE_CLASSES'] = extended

            elif isinstance(default_value, dict):
                defaults[key].update(value)

            else:  # pragma: nocover
                raise ValueError(f'Unable to extend `{key}` option.')

        installed_apps = defaults['INSTALLED_APPS']

        if app_name:

            if app_name not in installed_apps:
                installed_apps.append(app_name)

        else:
            dir_current = pytest_config.invocation_dir
            dir_tests = None

            app_name = dir_current.basename

            if app_name == 'tests':
                # Support certain module or function invocation tests dir as base (e.g. PyCharm behaviour).
                app_name = dir_current.parts()[-2].basename
                dir_tests = dir_current

            try:
                dir_tests = dir_current.listdir('tests')[0]

            except IndexError:
                pass

            if not dir_tests:
                # No `tests` subdir found. Let's try to deduce.

                app_name = None

                from setuptools import find_packages
                import py

                candidate_latest = ''
                candidates = []

                for package in find_packages(f'{dir_current}'):
                    # Consider only top level packages.
                    if not candidate_latest or not package.startswith(candidate_latest):
                        candidates.append(package)
                        candidate_latest = package

                for candidate in candidates:
                    dirs = py.path.local(candidate).listdir('tests')

                    if dirs:
                        app_name = candidate
                        dir_tests = dirs[0]
                        break

            if not app_name and not project_mode:
                raise Exception(
                    'Unable to deduce application name. '
                    'Check application package and `tests` directory exists. '
                    f'Current dir: {dir_current}')

            if app_name:
                installed_apps.append(app_name)

            if dir_tests:
                # Try to find and add an additional test app.
                dir_testapp_name = cls.DIR_TESTAPP
                dir_testapp = dir_tests.listdir(dir_testapp_name)

                if dir_testapp:
                    dir_testapp = dir_testapp[0]

                    testapp_name = f'{app_name}.tests.{dir_testapp_name}'

                    installed_apps.append(testapp_name)

                    if dir_testapp.listdir('urls.py'):
                        # Set customized `urls.py`.
                        defaults['ROOT_URLCONF'] = f'{testapp_name}.urls'

        djapp_options[cls.KEY_APP] = app_name

        defaults = hook(defaults)

        return defaults

    @classmethod
    def get_dict(cls) -> dict:
        """Returns current configuration as a dict."""
        return cls.get()[cls._prefix]



from typing import List, Dict

import pytest

from collections import defaultdict

if False:  # pragma: nocover
    from django.contrib.messages.storage.base import Message  # noqa


class DjangoappMessageStorage:

    def __init__(self):
        self.all: List['Message'] = []
        self.tags: Dict[str, List['Message']] = defaultdict(list)

    def __contains__(self, item):
        for msg in self.all:
            if item in f'{msg}':
                return True

        return False

    def __len__(self):
        return len(self.all)

    def add(self, level, message, extra_tags=''):
        from django.contrib.messages.storage.base import Message

        msg = Message(level, message, extra_tags)
        self.all.append(msg)
        self.tags[msg.level_tag].append(msg)


@pytest.fixture
def messages(monkeypatch) -> DjangoappMessageStorage:
    """Holds messages sent by Django Messages framework.

    Attributes:
        * all -- a list of all messages
        * tags -- messages dictionary indexed by tags

    Example::

        def test_this(messages):
            assert len(messages.all)
            assert len(messages.tags['error'])
            assert 'another error' in messages

    """
    messages = DjangoappMessageStorage()

    def add(request, level, message, extra_tags='', **kwargs):
        messages.add(level, message, extra_tags)

    monkeypatch.setattr('django.contrib.messages.api.add_message', add)

    return messages



from django.core.mail import send_mail


def test_mail_outbox(mail_outbox):

    send_mail(
        'Subject here', 'Here is the message.',
        'from@example.com', ['to@example.com'],
        fail_silently=False,
    )

    assert len(mail_outbox) == 1
    assert mail_outbox[0].subject == 'Subject here'



import pytest

def test_settings(settings):
    from django.conf import settings as _actual_settings

    assert not _actual_settings.DEBUG

    settings.update({
        'DEBUG': True,
        'SOMENEW': 'value',
    })

    assert settings.DEBUG
    assert settings.SOMENEW == 'value'

    assert _actual_settings.DEBUG
    assert _actual_settings.SOMENEW == 'value'

    settings.SOMENEW = 'other'

    assert settings.SOMENEW == 'other'

    assert _actual_settings.SOMENEW == 'other'


def test_context_manager(settings):
    from django.conf import settings as _actual_settings

    assert not _actual_settings.DEBUG

    with settings(SOME=1, DEBUG=True):
        assert settings.SOME == 1
        assert settings.DEBUG

        assert _actual_settings.SOME == 1
        assert _actual_settings.DEBUG

        settings.OTHER = 1
        assert settings.OTHER == 1
        assert _actual_settings.OTHER == 1

    assert not hasattr(settings, 'SOME')
    assert not hasattr(settings, 'OTHER')

    assert not _actual_settings.DEBUG
    assert not hasattr(_actual_settings, 'SOME')
    assert not hasattr(_actual_settings, 'OTHER')

    assert 'django.contrib.sites' in _actual_settings.INSTALLED_APPS
    assert 'django.contrib.sessions.middleware.SessionMiddleware' in _actual_settings.MIDDLEWARE
    assert 'django.contrib.sessions.middleware.SessionMiddleware' in _actual_settings.MIDDLEWARE_CLASSES
    assert 'dummy' in _actual_settings.DATABASES

    settings.DEBUG = False

    assert not _actual_settings.DEBUG

    settings.SOME = 2

    assert settings.SOME == 2

    assert _actual_settings.SOME == 2

    with settings(SOME='value', DEBUG=True):
        assert settings.SOME == 'value'
        assert settings.DEBUG

        assert _actual_settings.SOME == 'value'
        assert _actual_settings.DEBUG

    assert not settings.DEBUG
    assert not _actual_settings.DEBUG

    # SOME from before the context manager is now removed.
    assert not hasattr(settings, 'SOME')
    assert not hasattr(_actual_settings, 'SOME')

    with settings:
        settings.OTHER = 1
        assert settings.OTHER == 1
        assert _actual_settings.OTHER == 1

    assert not _actual_settings.DEBUG

    assert not hasattr(settings, 'SOME')
    assert not hasattr(settings, 'OTHER')

    assert not hasattr(_actual_settings, 'SOME')
    assert not hasattr(_actual_settings, 'OTHER')

    settings.OTHER = 1
    assert settings.OTHER == 1



import re
from typing import Union
from unittest import mock

import pytest
from django import VERSION
from django.template.base import Template
from django.template.context import Context, RenderContext

if False:  # pragma: nocover
    from django.contrib.auth.models import AbstractUser  # noqa
    from django.http import HttpRequest


RE_TAG_VALUES = re.compile('>([^<]+)<')


@pytest.fixture
def template_strip_tags():
    """Allows HTML tags strip from string.

    To be used with `template_render_tag` fixture to easy result assertions.

    Example::

        def test_this(template_strip_tags):
            stripped = template_strip_tags('<b>some</b>')

    :param html: HTML to strin tags from
    :param joiner: String to join tags contents. Default: |

    """
    def template_strip_tags_(html: str, joiner: str = '|') -> str:

        result = []
        for match in RE_TAG_VALUES.findall(html):
            match = match.strip()
            if match:
                result.append(match)

        return joiner.join(result)

    return template_strip_tags_


@pytest.fixture
def template_context(request_get, user_create):
    """Creates template context object.

    To be used with `template_render_tag` fixture.

    Example::

        def test_this(template_context):
            context = template_context({'somevar': 'someval'})

    :param context_dict: Template context. If not set empty context is used.

    :param request: Expects HttpRequest or string.
        String is used as a path for GET-request.

    :param current_app:

    :param user: User instance to associate request with.
        Defaults to a new anonymous user.
        If string is passed, it is considered to be


    """
    def template_context_(
        context_dict: dict = None,
        *,
        request: 'HttpRequest' = None,
        current_app: str = '',
        user: Union[str, 'AbstractUser'] = 'anonymous'
    ) -> Context:

        context_dict = context_dict or {}

        if user and isinstance(user, str):
            if user == 'anonymous':
                user = user_create(anonymous=True)
            else:
                user = user_create(attributes={'username': user})

        if not request or isinstance(request, str):
            request = request_get(request, user=user)

        context_updater = {
            'request': request,
        }

        if user:
            context_updater['user'] = user

        context_dict.update(context_updater)

        context = Context(context_dict)
        contribute_to_context(context, current_app)

        return context

    return template_context_


@pytest.fixture
def template_render_tag():
    """Renders a template tag from a given library by its name.

    Example::

        def test_this(template_render_tag):
            rendered = template_render_tag('library_name', 'mytag arg1 arg2')

    :param library: Template tags library name to load tag from.

    :param tag_str: Tag string itself. As used in templates, but without {% %}.

    :param context: Template context object. If not set,
        empty context object is used.

    """
    def template_render_tag_(library: str, tag_str: str, context: Union[dict, Context] = None) -> str:
        context = context or {}

        if not isinstance(context, Context):
            context = Context(context)

        string = f'{{% load {library} %}}{{% {tag_str} %}}'

        template = Template(string)
        contribute_to_context(context, template=template)

        if VERSION >= (1, 11):
            # Prevent "TypeError: 'NoneType' object is not iterable" in  get_exception_info
            template.nodelist[1].token.position = (0, 0)

        return template.render(context)

    return template_render_tag_


def contribute_to_context(context: Context, current_app: str = '', template: Template = None):
    template = template or mock.MagicMock()

    if VERSION >= (1, 8):
        template.engine.string_if_invalid = ''

    context.template = template

    if VERSION >= (1, 11):
        context.render_context = RenderContext()

    if VERSION >= (1, 10):
        match = mock.MagicMock()
        match.app_name = current_app
        context.resolver_match = match

    else:
        context._current_app = current_app



import pytest

from django import VERSION


@pytest.fixture()
def check_migrations(command_makemigrations):
    """Check if migrations are up to date (migration files exist for current models' state).

    Example::

        def test_this(check_migrations):
            result = check_migrations()

    :param app: Application name to check migrations for.

    """
    def check_migrations_(app: str = ''):
        if VERSION < (2, 0):
            raise Exception('Django 2.0+ required for checking migrations')

        try:
            command_makemigrations(app=app, args=['--check', '--dry-run'])

        except SystemExit:
            return False

        return True

    return check_migrations_



import pytest
from django.test import RequestFactory, Client

try:
    from django.urls import reverse

except ImportError:  # Django < 2.0
    from django.core.urlresolvers import reverse


if False:  # pragma: nocover
    from django.contrib.auth.models import AbstractUser  # noqa
    from django.http import HttpRequest


def _contribute_ajax(headers, flag):
    if flag:
        headers['HTTP_X_REQUESTED_WITH'] = 'XMLHttpRequest'


class DjangoappRequestFactory(RequestFactory):

    def __init__(self, ajax=False, json=False, **defaults):
        _contribute_ajax(defaults, ajax)
        super(DjangoappRequestFactory, self).__init__(**defaults)
        self._json = json

    def _encode_json(self, data, content_type):
        if self._json:
            content_type = 'application/json'
        return super(DjangoappRequestFactory, self)._encode_json(data, content_type)

    def _encode_data(self, data, content_type):
        if self._json:
            content_type = 'application/json'
        return super(DjangoappRequestFactory, self)._encode_data(data, content_type)

    def generic(self, method, path, *args, **kwargs):

        if isinstance(path, tuple):
            path = reverse(path[0], kwargs=path[1])

        return super(DjangoappRequestFactory, self).generic(method, path, *args, **kwargs)


class DjagoappClient(Client, DjangoappRequestFactory):

    def __init__(
            self,
            *,
            ajax: bool = False,
            user: 'AbstractUser' = None,
            enforce_csrf_checks: bool = False,
            raise_exceptions: bool = True,
            json: bool = False,
            **defaults
    ):
        _contribute_ajax(defaults, ajax)

        super(DjagoappClient, self).__init__(enforce_csrf_checks, json=json, **defaults)

        if user:
            assert hasattr(user, 'password_plain'), (
                'Request client expects a user created with `user` or `user_create` fixtures.')

            logged_in = self.login(username=user.username, password=user.password_plain)

        else:
            logged_in = None

        self.user = user
        self.user_logged_in = logged_in
        self._raise_exc = raise_exceptions

    def store_exc_info(self, **kwargs):
        if self._raise_exc:
            super(DjagoappClient, self).store_exc_info(**kwargs)


@pytest.fixture
def request_factory():
    """Fixture allowing request object generation.

    Example::

        def test_this(request_factory):
            factory = request_factory()

    :param kwargs:

    """
    def request_factory_(**kwargs) -> DjangoappRequestFactory:
        return DjangoappRequestFactory(**kwargs)

    return request_factory_


@pytest.fixture
def request_get(request_factory):
    """Fixture allowing GET request object generation.

    Example::

        def test_this(request_get):
            request = request_get('/some')


    :param path:

    :param user: User making this request.

    :param ajax: Make AJAX (XMLHttpRequest) request.

    :param kwargs: Additional arguments for .get() method.

    """
    def request_get_(path: str = None, *, user: 'AbstractUser' = None, ajax: bool = False, **kwargs) -> 'HttpRequest':

        path = path or '/'
        request = request_factory(ajax=ajax).get(path, **kwargs)
        if user:
            request.user = user
        return request

    return request_get_


@pytest.fixture
def request_post(request_factory):
    """Fixture allowing POST request object generation.

    Example::

        def test_this(request_post):
            request = request_post('/some', {'a': 'b'})


    :param path:

    :param data: Data to post.

    :param user: User making this request.

    :param ajax: Make AJAX (XMLHttpRequest) request.

    :param kwargs: Additional arguments for .post() method.

    """
    def request_post_(
        path: str = None,
        data: dict = None,
        *,
        user: 'AbstractUser' = None,
        ajax: bool = False,
        **kwargs
    ) -> 'HttpRequest':

        path = path or '/'
        request = request_factory(ajax=ajax).post(path, data, **kwargs)
        if user:
            request.user = user
        return request

    return request_post_


@pytest.fixture
def request_client():
    """Fixture allowing test client object generation.

    Example::

        def test_this(request_client):

            client = request_client()

            response = client.get(
                ('someview', {'somearg': 'one', 'otherarg': 33})
            ).content

            ...

            ajax_client = request_client(ajax=True)
            ...


    :param ajax: Make AJAX (XMLHttpRequest) requests.

    :param user: User to perform queries from.

    :param raise_exceptions: Do not allow Django technical exception handlers
        to catch exceptions issued by views, propagate them instead.

    :param json: Encode data as JSON.

        .. warning:: To be used with Django 2.1+

    :param kwargs: Additional arguments for test client initialization.

    """
    def request_client_(
        *,
        ajax: bool = False,
        user: 'AbstractUser' = None,
        raise_exceptions: bool = True,
        json: bool = False,
        **kwargs
    ) -> DjagoappClient:

        return DjagoappClient(
            ajax=ajax,
            user=user,
            raise_exceptions=raise_exceptions,
            json=json,
            **kwargs
        )

    return request_client_



import re
import pytest

from django import VERSION


@pytest.mark.skipif(VERSION < (2, 0), reason='Django 2.0+ required')
def test_check_migrations_missing(check_migrations):
    result = check_migrations('testapp_missing_migrations')
    assert result is False, 'Migrations should be missing in testapp_missing_migrations'


@pytest.mark.skipif(VERSION < (2, 0), reason='Django 2.0+ required')
def test_check_migrations_ok(check_migrations):
    result = check_migrations()
    assert result, (
        "ERROR: Migrations check failed! Models' changes not migrated, "
        "please run './manage.py makemigrations' to solve the issue!"
    )


@pytest.mark.skipif(VERSION >= (2, 0), reason='Testing Django 1.9- required')
def test_old_django(check_migrations):
    with pytest.raises(Exception, match=re.escape('Django 2.0+ required for checking migrations')):
        check_migrations()



import sys

import pytest

from pytest_djangoapp import configure_djangoapp_plugin
from pytest_djangoapp.configuration import Configuration


@pytest.mark.skipif(sys.version_info > (2, 0), reason='teardown without setup should pass')
def test_teardown_without_setup():
    assert True


@pytest.mark.skipif(sys.version_info < (3, 0), reason='whole project test are for Py 3')
def test_whole_project_testing(request):
    configure_djangoapp_plugin('pytest_djangoapp.tests.settings_project')
    settings_dict = Configuration.get_combined(request.config)
    assert '_' not in settings_dict
    assert 'SOME_ATTR' in settings_dict
    assert 'some_function' not in settings_dict




def test_messages(messages):

    assert not messages.all

    from django.contrib import messages as djmessages

    djmessages.error(None, 'this is error')
    djmessages.error(None, 'ещё ошибка')
    djmessages.info(None, 'some info')

    assert len(messages) == 3
    assert len(messages.tags['error']) == 2
    assert len(messages.tags['info']) == 1

    assert 'ещё ошибка' in messages
    assert 'other' not in messages




def test_template_context(template_context, user_create):
    assert template_context({'somevar': 'someval'})

    context = template_context({'a': 'b'}, user='aname')
    assert context.get('user').username == 'aname'

    auser = user_create(attributes={'username': 'xx'})
    context = template_context({'a': 'b'}, user=auser)
    assert context.get('user').username == 'xx'


def test_template_render_tag(template_render_tag):
    rendered = template_render_tag('static', 'static "some.jpg"')
    assert rendered == '/static/some.jpg'


def test_this_template_strip_tags(template_strip_tags):
    stripped = template_strip_tags('<b>some</b>')
    assert stripped == 'some'



from functools import partial
from typing import Union, Callable

from .configuration import Configuration


def configure_djangoapp_plugin(
    settings: Union[str, dict] = None,
    *,
    app_name: str = None,
    admin_contrib: bool = False,
    settings_hook: Callable = None,
    migrate: bool = True,
    **kwargs
) -> str:
    """Configures djangoapp pytest plugin.

    :param settings: Django project settings to override defaults.
        To get settings from a module (for testing an entire project, not just one app)
        use a string with the module name.

    :param app_name: Name of your application you write tests for.
        If not set ``djangoapp`` will try to guess (``tests`` dir needs to be inside application directory).

    :param admin_contrib: Setup Django to test Admin contrib related parts.

    :param settings_hook: Allows setting a function to get resulting settings.

        Function must accept settings dictionary, and return resulting settings dictionary.

        .. code-block:: python

            def hook_func(settings):
                return settings

    :param bool migrate: Allows applying or skipping migration applying process. 
        Skipping could be useful for testing applications with many migrations.

    :param kwargs: Additional arguments.

        Use `extend_` prefix to extend default configuration.

        .. code:: python

            configure_djangoapp_plugin(

                extend_INSTALLED_APPS=[
                    'django.contrib.auth',
                    'django.contrib.contenttypes',
                    'django.contrib.sessions'
                ],
                extend_DATABASES={
                    'dummy': {
                        'ENGINE': 'django.db.backends.sqlite3',
                        'NAME': ':memory:',
                    },
                },
            )

    :returns: Plugin name to add into ``pytest_plugins`` module attribute of ``conftest.py``.

    """
    try:
        import django

    except ImportError:  # pragma: nocover
        raise Exception('Django is not available in test environment. Please install it.')

    if isinstance(settings, str):
        # Considering a whole project testing mode.
        settings_hook = partial(update_settings_from_module, module=settings)
        settings = {}

    Configuration.set(
        settings_dict=settings,
        app_name=app_name,
        admin_contrib=admin_contrib,
        settings_hook=settings_hook,
        migrate=migrate,
        **kwargs
    )

    return 'pytest_djangoapp.plugin'


def update_settings_from_module(settings: dict, *, module: str) -> dict:
    """Updates a given settings dict from a module denoted by a dotted path.

    :param settings:
    :param module:

    """
    from importlib import import_module

    settings_module = import_module(module)

    settings_dict = {
        key: value
        for key, value in settings_module.__dict__.items()
        if key.upper() == key and not key.startswith('_')
    }
    settings.update(settings_dict)

    return settings



from django import VERSION


def get_urlpatterns(patterns_list: list) -> list:
    """Returns object suitable to use as urlpatterns in `urls.py`.

    Example::
        urlpatterns = get_urlpatterns([
            url(r'^index/$', index, name='index'),
        ])

    :param patterns_list:

    """
    if VERSION >= (1, 9):
        return patterns_list

    from django.conf.urls import patterns

    return patterns('', *patterns_list)



from typing import Any

import pytest
from django.conf import settings as django_settings


_UNSET = set()


@pytest.fixture()
def settings() -> 'SettingsProxy':
    """Fixture allowing to temporarily override project settings.

    Example::

        def test_this(settings):
            # By one.
            settings.DEBUG = True

            # Batch.
            settings.update({
                'DEBUG': True,
                'ROOT_URLCONF': 'module.urls',
            })

            # Context manager.
            with settings(DEBUG=True):
                ...

    """
    proxy = SettingsProxy()

    yield proxy

    proxy.restore_initial()


_PROXY_SETTING_ATTRS = '_settings', '_overridden'


class SettingsProxy:

    def __init__(self):
        self._settings = django_settings
        self._overridden = {}

    def __setattr__(self, name: str, value: Any):

        if name in _PROXY_SETTING_ATTRS:
            return super(SettingsProxy, self).__setattr__(name, value)

        self._set(name, value)

    def __getattr__(self, name: str) -> Any:

        if name in _PROXY_SETTING_ATTRS:
            return super(SettingsProxy, self).__getattr__(name)

        try:
            return getattr(self._overridden, name)
        except AttributeError:
            return getattr(self._settings, name)

    def __call__(self, **kwargs) -> 'SettingsProxy':
        # Aid context manager mode.
        do_set = self._set

        for key, val in kwargs.items():
            do_set(key, val)

        return self

    def __enter__(self) -> 'SettingsProxy':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.restore_initial()

    def _set(self, name: str, value: Any):
        settings = self._settings
        overridden = self._overridden

        current_val = getattr(settings, name, _UNSET)

        if name not in overridden:
            # Handle subsequent changes.
            overridden[name] = current_val

        setattr(settings, name, value)

    def update(self, settings: dict):
        """Mass update settings.

        :param settings:

        """
        for name, value in settings.items():
            self._set(name, value)

    def restore_initial(self):
        """Restore initial settings."""

        settings = self._settings
        unset = _UNSET

        for key, val in self._overridden.items():
            if val is unset:
                # wipe
                delattr(settings, key)

            else:
                # restore
                setattr(settings, key, val)



from unittest.mock import MagicMock

import pytest
from django import VERSION


@pytest.mark.skipif(VERSION < (4, 0), reason='For Django 4.0+')
def test_liveserver(liveserver, liveclient):

    with liveserver(host='localhost', port=33445) as server:
        assert server.url == 'http://localhost:33445'


@pytest.mark.skipif(VERSION < (4, 0), reason='For Django 4.0+')
def test_selenium(liveserver, liveclient, monkeypatch):

    monkeypatch.setattr('django.test.selenium.SeleniumTestCaseBase', MagicMock())

    with liveserver() as server:

        with liveclient('selenium', browser='firefox') as client:
            selenium = client.handle

            selenium.get(server.url)
            assert 'MagicMock' in f"{selenium.find_element('tag name', 'h1').text}"
